"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import os
import shutil
import unittest
from kudubot.tests.helpers.DummyConnection import DummyConnection
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler
from kudubot.tests.helpers.DummyService import DummyService, DummyServiceWithValidDependency
from kudubot.tests.helpers.test_config import generate_test_environment, clean_up_test_environment


class UnitTests(unittest.TestCase):
    """
    Tests the GlobalConfigHandler class
    """

    def setUp(self):
        """
        Creates a restore point for the class variables of the GlobalConfigHandler and sets these values
        to ones that make sense for the unit tests
        :return: None
        """
        self.config_handler = generate_test_environment()

    def tearDown(self):
        """
        Restores the class variables and deletes any temporary directories and files
        :return: None
        """
        clean_up_test_environment()

    def test_generating_new_config(self):
        """
        Tests if the configuration generation works as intended
        :return: None
        """
        clean_up_test_environment()

        self.config_handler.generate_configuration(True)
        self.validate_config_directory()
        self.config_handler.generate_configuration(True)
        self.validate_config_directory()
        self.config_handler.generate_configuration(False)
        self.validate_config_directory()

    def validate_config_directory(self):
        """
        Validates a configuration directory
        :return: None
        """
        self.assertTrue(self.config_handler.validate_config_directory())
        self.assertTrue(os.path.isdir("kudu-test"))
        self.assertTrue(os.path.isfile(os.path.join("kudu-test", "services.conf")))
        self.assertTrue(os.path.isfile(os.path.join("kudu-test", "connections.conf")))
        self.assertTrue(os.path.isdir(os.path.join("kudu-test", "connection_config")))
        self.assertTrue(os.path.isdir(os.path.join("kudu-test", "data")))

    def assure_invalid_config_directory(self):
        """
        Makes sure that the current configuration directory is invalid
        :return: None
        """
        self.assertFalse(self.config_handler.validate_config_directory())

    def test_directory_validation(self):
        """
        Tests if the global config validation works correctly and finds errors in the config structure
        :return: None
        """
        clean_up_test_environment()

        self.assure_invalid_config_directory()

        for element in ["services.conf", "connections.conf", "data", "connection_config"]:

            self.config_handler.generate_configuration(True)
            path = os.path.join("kudu-test", element)

            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)

            self.assure_invalid_config_directory()

    def test_connection_loading(self):
        """
        Tests the loading of connections

        :return: None
        """
        connections = self.config_handler.load_connections()
        self.assertEqual([], connections)

        with open(os.path.join("kudu-test", "connections.conf"), 'w') as f:
            f.write("from kudubot.tests.helpers.DummyConnection import DummyConnection")

        self.assertEqual(self.config_handler.load_connections()[0], DummyConnection)

    def test_service_loading(self):
        """
        Tests the loading of services

        :return: None
        """
        services = self.config_handler.load_services()

        self.assertEqual([], services)

        with open(os.path.join("kudu-test", "services.conf"), 'w') as f:
            f.write("from kudubot.tests.helpers.DummyService import DummyService")

        self.assertEqual(self.config_handler.load_services()[0], DummyService)

    def test_importing(self):
        """
        Tests the String import handler method

        :return: None
        """
        os_import = self.config_handler.__handle_import_statement__("import os")
        dict_import = self.config_handler.__handle_import_statement__("from typing import Dict")
        from typing import Dict

        self.assertEqual(os, os_import)
        self.assertEqual(dict_import, Dict)

    def test_service_dependency_imports(self):
        """
        Tests the dependency resolution of the services.

        :return: None
        """
        # First, test service having a valid dependency
        with open(os.path.join("kudu-test", "services.conf"), 'w') as f:
            f.write("from kudubot.tests.helpers.DummyService import DummyServiceWithValidDependency")
        services = self.config_handler.load_services()
        self.assertEqual(services, [DummyServiceWithValidDependency])

        # Now test unresolved dependency
        with open(os.path.join("kudu-test", "services.conf"), 'w') as f:
            f.write("from kudubot.tests.helpers.DummyService import DummyServiceWithInvalidDependency")
        services = self.config_handler.load_services()
        self.assertEqual(services, [])

    def test_duplicate_removal(self):
        """
        Tests the duplicate removal method

        :return: None
        """
        # noinspection PyDecorator
        def other():
            return "other"

        second_dummy = DummyConnection([], self.config_handler)
        second_dummy.define_identifier = other

        connections = [second_dummy, DummyConnection, DummyConnection]
        # noinspection PyTypeChecker
        new_connections = self.config_handler.__remove_duplicate_services_or_connections__(connections)
        self.assertEqual(new_connections, [second_dummy, DummyConnection])

    def test_loading_invalid_classes(self):

        with open(os.path.join("kudu-test", "connections.conf"), 'w') as f:
            f.write("from kudubot.tests.helpers.DummyService import DummyService")

        self.assertEqual(self.config_handler.load_connections(), [])

        with open(os.path.join("kudu-test", "connections.conf"), 'w') as f:
            f.write("from kudubot.doesnotexists import AClass")

        self.assertEqual(self.config_handler.load_connections(), [])



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import sqlite3
from typing import Dict, List
from kudubot.entities.Message import Message
from kudubot.services.Service import Service
from kudubot.connections.Connection import Connection


# noinspection PyAbstractClass,SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
class MultiLanguageService(Service):
    """
    Service Extension class that enables support for multiple languages
    """

    def __init__(self, connection: Connection):
        """
        In addition to the normal initialization of a Service, this service initializes
        its database which stores information about a user's language preferences
        :param connection: The connection which is used to communicate
        """
        super().__init__(connection)
        self.connection.db.execute("CREATE TABLE IF NOT EXISTS language_preferences ("
                                   "user_id INTEGER CONSTRAINT constraint_name PRIMARY KEY, "
                                   "lang_pref VARCHAR(255) NOT NULL,"
                                   "user_initiated INT NOT NULL"
                                   ")")
        self.connection.db.commit()

    def store_language_preference(self, user: int, language: str, user_initiated: bool = False):
        """
        Stores the language preference for a user in the sqlite database
        :param user: The user for which to store the preference
        :param language: The language to store
        :param user_initiated: Flag that should be set whenever the user manually requests a language store
        :return: None
        """
        fetch = self.connection.db.execute("SELECT user_initiated FROM language_preferences WHERE user_id=?",
                                           (user,)).fetchall()

        was_user_initiated = len(fetch) > 0 and fetch[0][0]
        if was_user_initiated and not user_initiated:
            return  # Don't overwrite user defined language

        else:
            self.connection.db.execute("INSERT OR REPLACE INTO language_preferences "
                                       "(user_id, lang_pref, user_initiated) VALUES (?,?,?)",
                                       (user, language, user_initiated))
            self.connection.db.commit()

    def get_language_preference(self, user: int, default: str = "en", db: sqlite3.Connection=None) -> str:
        """
        Retrieves a language from the user's preferences in the database

        :param user: The user to check the language preference for
        :param default: A default language value used in case no entry was found
        :param db: Optionally defines which database connection to use (necessary for access from other thread)
        :return: The language preferred by the user
        """
        db = db if db is not None else self.connection.db
        result = db.execute("SELECT lang_pref FROM language_preferences WHERE user_id=?", (user,)).fetchall()
        return default if len(result) == 0 else result[0][0]

    def determine_language(self, message: Message) -> str:
        """
        Checks a message object for any language indicators to determine
        in which language to reply with

        :param message: The message to analyze
        :return: The language key
        """
        raise NotImplementedError()

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        Defines the dictionary with which the text is translated in the translate() method.
        This should be in the form of a dictionary like this:

        { key: {"language": "text_in_language", ...}, ... }

        Keep in mind that every instance of the 'key' value is replaced while translating

        :return: The dictionary to create translations with
        """
        raise NotImplementedError

    def supported_languages(self) -> List[str]:
        """
        :return: A list of languages supported by the Service
        """

        supported = []

        dictionary = self.define_language_text()
        for token in dictionary:
            for language in dictionary[token]:
                if language not in supported:
                    supported.append(language)
        return supported

    # noinspection PyMethodMayBeStatic
    def define_fallback_language(self) -> str:
        """
        Defines a fallback language in case a language is not implemented for a key.

        :return: By default, the language "en" is returned
        """
        return "en"

    def translate(self, text: str, language: str, translation_dict: Dict[str, Dict[str, str]]=None) -> str:
        """
        Translates text using the service's dictionary in the specified language

        :param text: The text to translate
        :param language: The language to translate into
        :param translation_dict: Can be specified to determine a custom dictionary
        :return: The translated text
        """

        translated = text
        language_text = self.define_language_text() if translation_dict is None else dict(translation_dict)

        for key in language_text:
            try:
                translated = translated.replace(key, language_text[key][language])
            except KeyError:
                translated = translated.replace(key, language_text[key][self.define_fallback_language()])

        return translated

    def reply_translated(self, title: str, body: str, message: Message):
        """
        Provides a helper method that streamlines the process of replying to a message. Very useful
        for Services that send a reply immediately to cut down on clutter in the code

        In addition to the standard reply method, this method translates the text prior to sending it

        :param title: The title of the message to send
        :param body: The body of the message to send
        :param message: The message to reply to
        :return: None
        """
        language = self.get_language_preference(message.get_direct_response_contact().database_id)
        self.reply(self.translate(title, language), self.translate(body, language), message)

    def is_applicable_to(self, message: Message):
        body = message.message_body.lower().strip()
        return body.startswith("/language ") or body.startswith("/sprache ") or body in ["/language", "/sprache"]

    def handle_message(self, message: Message):
        """
        Analyzes a message for the language used and stores that language value in the database as a preference of
        the user.
        Also implements the /language command which allows a user to view and change the current language

        :param message: The message to analyze
        :return: None
        """

        dictionary = {"@title": {"en": "Language Change",
                                 "de": "Sprachwechsel"},
                      "@success_message": {"en": "Successfully changed language to",
                                           "de": "Sprache erfolgreich geändert zu"},
                      "@fail_message": {"en": "Failed to switch to language",
                                        "de": "Konnte nicht Sprache wechseln zu"}}

        command_keywords = ["/language", "/sprache"]
        aliases = {"en": ["en", "english", "englisch"],
                   "de": ["de", "deutsch", "german"]}

        params = message.message_body.lower().split(" ")
        user_id = message.get_direct_response_contact().database_id

        if len(params) == 1 and params[0] in command_keywords:
            language = self.get_language_preference(user_id, "en")
            self.reply(self.translate("@title", language, dictionary),
                       language, message)
            return  # Important, makes overriding methods quit when help message is applicable

        if len(params) == 2 and params[0] in command_keywords:
            found_language = False
            for key in aliases:
                for alias in aliases[key]:
                    if alias == params[1]:
                        found_language = True
                        self.store_language_preference(user_id, key, True)
                        break

            language = self.get_language_preference(user_id, "en")  # the new language will already be stored
            title = self.translate("@title", language, dictionary)

            if found_language:
                self.reply(title, self.translate("@success_message: " + language, language, dictionary), message)
            else:
                self.reply(title, self.translate("@fail_message: " + params[1], language, dictionary), message)

            return  # Important, makes overriding methods quit when help message is applicable

        else:

            try:
                language = self.determine_language(message)
                self.store_language_preference(message.get_direct_response_contact().database_id, language)
            except NotImplementedError:
                pass



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from kudubot.entities.Message import Message
from kudubot.services.MultiLanguageService import MultiLanguageService


# noinspection PyAbstractClass
class HelperService(MultiLanguageService):
    """
    Service extension that allows for the automatic sending of help and syntax messages.
    Provides support for multiple languages
    """

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message for the Service

        :param language: The language in which to get the help message in
        :return: The help message in the specified language
        """
        raise NotImplementedError()

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the syntax description for this service.

        :param language: The language in which to get the syntax description in
        :return: The syntax description in the specified language
        """
        raise NotImplementedError()

    # noinspection PyUnusedLocal
    def define_command_name(self, language: str) -> str:
        """
        Defines the command name used to call this Service

        :param language: The language for the command name, for supporting different command names for
                         different languages
        :return: The command name for this service. Defaults to a forward slash and the Service's identifier
        """
        return "/" + self.define_identifier()

    def handle_message(self, message: Message):
        """
        Handles the help message sending. Checks if a message qualifies for a help message and then sends
        messages accordingly.

        Subclasses of the HelperService should call this method using super()

        :param message: The message to handle
        :return: None
        """
        super().handle_message(message)
        if MultiLanguageService.is_applicable_to(self, message):
            return

        body = message.message_body

        try:
            language = self.determine_language(message)
        except NotImplementedError:
            language = "en"

        dictionary = {
            "@help_message_title": {"en": "Help Message for " + self.identifier,
                                    "de": "Hilfnachricht für " + self.identifier},
            "@syntax_message_title": {"en": "Syntax Message for " + self.identifier,
                                      "de": "Syntaxnachricht für " + self.identifier}
        }
        help_keywords = ["help", "hilfe"]
        syntax_keywords = ["syntax"]

        if body.startswith(self.define_command_name(language)):
            body = body.split(self.define_command_name(language), 1)[1].strip()

            if body in help_keywords:
                self.reply(self.translate("@help_message_title", language, dictionary),
                           self.define_help_message(language), message)
                return
            elif body in syntax_keywords:
                self.reply(self.translate("@syntax_message_title", language, dictionary),
                           self.define_syntax_description(language), message)
                return

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if the message is applicable to the service by checking if the command name is followed by
        the terms 'help' or 'syntax'.

        :param message: The message to analyze
        :return: True if the message is applicable, False otherwise
        """
        if super().is_applicable_to(message):
            return True

        language = self.determine_language(message)
        command = self.define_command_name(language).lower()
        dictionary = {
            "@help_command": {"en": "help",
                              "de": "hilfe"},
            "@syntax_command": {"en": "syntax",
                                "de": "syntax"}
        }

        return message.message_body.lower() in [
            command + self.translate(" @help_command", language, dictionary),
            command + self.translate(" @syntax_command", language, dictionary)
        ]

    def is_applicable_to_without_help_or_syntax(self, message: Message) -> bool:
        """
        Checks if the message applies to anything beside the 'help' or 'syntax' commands

        :param message: The message to analyze
        :return: True if the message is applicable to something else, False otherwise
        """
        return self.is_applicable_to(message) and not HelperService.is_applicable_to(self, message)



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import os
import stat
from typing import List
from subprocess import Popen


# noinspection PyUnusedLocal
def build(service_name: str, service_dir: str, source_file: str, destination_file: str):
    """
    Builds an executable file for an external service.

    :param service_name: The name of the service
    :param service_dir: The root directory of the external service
    :param source_file: The main source file used for compilation
    :param destination_file: The destination executable file
    :return: None
    """

    current_dir = os.getcwd()

    if source_file.endswith(".rs"):  # Rust
        os.chdir(service_dir)
        run_safe_popen(["cargo", "build", "--release"])
        safe_move(os.path.join("target", "release", service_name), os.path.join("build", service_name))

    os.chdir(current_dir)


def run_safe_popen(command):
    """
    Runs a Popen command and makes sure to catch any exceptions that may occur.

    :param command: The command to run
    :return: None
    """
    try:
        Popen(command).wait()
    except BaseException as e:
        print(e)


def safe_move(source: str, destination: str):
    """
    Moves a file to a new location, but makes sure that the move is even possible.

    :param source: The source file
    :param destination: The destination file
    :return: None
    """
    if os.path.isfile(source) and os.path.isdir(os.path.dirname(destination)):
        os.rename(source, destination)


def build_external(move_to: str = "") -> List[str]:
    """
    Builds all external services

    :param move_to: If specified, all built files are moved to that directory
    :return: A list of all generated executable files
    """

    external_dir = os.path.join("kudubot", "services", "external")
    built_executables = []

    for service in os.listdir(external_dir):

        service_dir = os.path.join(external_dir, service)
        if not os.path.isdir(service_dir) or service == "__pycache__":
            continue

        src = os.path.join(service_dir, "src")
        build_dir = os.path.join(service_dir, "build")

        if not os.path.isdir(build_dir):
            os.makedirs(build_dir)

        for source_file in os.listdir(src):

            source_path = os.path.join(src, source_file)
            destination_path = os.path.join(build_dir, service)

            if os.path.isfile(source_path) and source_file.lower().startswith("main"):
                build(service, service_dir, source_path, destination_path)

                if os.path.isfile(destination_path):
                    st = os.stat(destination_path)
                    os.chmod(destination_path, st.st_mode | stat.S_IEXEC)  # Make executable
                    built_executables.append(destination_path)

    if move_to != "":
        for executable in built_executables:
            os.rename(executable, os.path.join(move_to, os.path.basename(executable)))

    return built_executables



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import logging
from typing import List
from kudubot.entities.Message import Message


class Service(object):
    """
    A class that defines how a chat bot service integrates with a Connection
    """

    def __init__(self, connection  # Connection  (Can't import due to circular imports)
                 ):
        """
        Initializes the Service using a specified Connection

        :param connection: The connection used by this service
        """
        self.connection = connection
        self.identifier = self.define_identifier()
        self.requires = self.define_requirements()
        self.logger = logging.getLogger("kudubot.services.Service." + self.identifier)
        self.init()

    # noinspection PyMethodMayBeStatic
    def init(self):
        """
        Helper method that runs after the initialization of the Service object. Can be used for
        anything, but normal use cases would include initializing a database table or
        starting a background thread

        :return: None
        """
        pass

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if the Service is applicable to a given message

        :param message: The message to check
        :return: True if applicable, else False
        """
        raise NotImplementedError()

    def handle_message(self, message: Message):
        """
        Handles the message, provided this service is applicable to it

        :param message: The message to process
        :return: None
        """
        raise NotImplementedError()

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the unique identifier for the service

        :return: The Service's identifier.
        """
        raise NotImplementedError()

    @staticmethod
    def define_requirements() -> List[str]:
        """
        Defines the requirements for the service

        :return: The required services for this Service.
        """
        raise NotImplementedError()

    def reply(self, title: str, body: str, message: Message):
        """
        Provides a helper method that streamlines the process of replying to a message. Very useful
        for Services that send a reply immediately to cut down on clutter in the code

        :param title: The title of the message to send
        :param body: The body of the message to send
        :param message: The message to reply to
        :return: None
        """
        reply_message = Message(title, body, message.get_direct_response_contact(), self.connection.user_contact)
        self.connection.send_message(reply_message)



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from typing import List
from kudubot.services.ExternalService import ExternalService


class HelloRustService(ExternalService):
    """
    A service that responds with 'Hello!' if you send it 'Hello Rust!'
    More a proof of concept than anything.
    """

    def define_executable_file_url(self):
        return self.resolve_github_release_asset_url("namboy94", "kudubot", "hello_rust")

    def define_executable_command(self) -> List[str]:
        return []

    @staticmethod
    def define_identifier() -> str:
        return "hello_rust"



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

"""
The metadata is stored here. It can be used by any other module in this project this way, most
notably by the setup.py file
"""

version = "0.14.2"  # pragma: no cover
"""
The current version of the program
"""

sentry_dsn = "https://f8dc6c1c49944cbdb5215e563e83cc78:" \
             "703626a5a0714e58b8c60947b949c524@sentry.io/160743"  # pragma: no cover
"""  # pragma: no cover
The DSN used for Sentry Error Logging
"""



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import os
import shutil
import time
import unittest

from kudubot.config.GlobalConfigHandler import GlobalConfigHandler
from kudubot.connections.Connection import Connection
from kudubot.entities.Message import Message
from kudubot.tests.helpers.DummyConnection import DummyConnection
from kudubot.tests.helpers.DummyService import DummyService
from kudubot.tests.helpers.test_config import generate_test_environment, clean_up_test_environment
from kudubot.users.AddressBook import Contact


class UnitTests(unittest.TestCase):
    """
    Class that tests the Connection class
    """

    def setUp(self):
        """
        :return: None
        """
        self.config_handler = generate_test_environment()
        self.connection = DummyConnection([], self.config_handler)

    def tearDown(self):
        """
        :return: None
        """
        clean_up_test_environment()

    def test_abstract_methods(self):
        """
        Tests if the methods of the connection class are abstract
        :return: None
        """
        for method in [(Connection.define_user_contact, 0),
                       (Connection.define_user_contact, 0),
                       (Connection.listen, 0),
                       (Connection.send_message, 1),
                       (Connection.send_image_message, 3),
                       (Connection.send_audio_message, 3),
                       (Connection.send_video_message, 3),
                       (Connection.generate_configuration, 0),
                       (Connection.load_config, 0),
                       (Connection.define_identifier, -1)]:
            try:

                # The self.connection arguments emulate the self parameter
                if method[1] == -1:
                    method[0]()
                if method[1] == 0:
                    method[0](self.connection)
                elif method[1] == 1:
                    method[0](self.connection, None)
                elif method[1] == 2:
                    method[0](self.connection, None, None)
                elif method[1] == 3:
                    method[0](self.connection, None, None, None)
                self.fail()
            except NotImplementedError:
                pass

    def test_daemon_thread_start(self):
        """
        Tests if the daemon thread is started correctly
        :return: None
        """
        class ThreadLessDummyConnection(DummyConnection):
            def listen(self):
                time.sleep(1)

        t = ThreadLessDummyConnection([], self.config_handler).listen_in_separate_thread()

        self.assertTrue(t.is_alive())
        while t.is_alive():
            pass
        self.assertFalse(t.is_alive())

    # noinspection PyMethodMayBeStatic
    def test_processing_message(self):
        """
        Tests if the connection correctly processes a message using the services
        :return: None
        """

        class AlwaysApplicableDummyService(DummyService):
            def is_applicable_to(self, message: Message):
                return True

        user = Contact(1, "1", "1")

        connection = DummyConnection([DummyService], self.config_handler)
        connection.apply_services(Message("Test", "Body", user, user))

        connection = DummyConnection([AlwaysApplicableDummyService], self.config_handler)
        connection.apply_services(Message("Test", "Body", user, user), True)
        connection.apply_services(Message("Test", "Body", user, user), False)

        connection.db.close()



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import sys
import raven
import logging
import argparse
from kudubot.metadata import version, sentry_dsn
from kudubot.exceptions import InvalidConfigException
from kudubot.connections.Connection import Connection
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


def main():  # pragma: no cover
    """
    The Main Method of the Program that starts the Connection Listener in accordance with the
    command line arguments

    :return: None
    """

    try:
        args = parse_args()

        # noinspection PyUnresolvedReferences
        if args.debug:
            logging.basicConfig(level=logging.DEBUG)
        elif args.verbose:
            logging.basicConfig(level=logging.INFO)

        # noinspection PyUnresolvedReferences
        connection = initialize_connection(args.connection.lower(), args.config)

        connection.listen()
    except Exception as e:
        sentry = raven.Client(dsn=sentry_dsn, release=version)
        sentry.captureException()
        raise e
    except KeyboardInterrupt:
        print("\nBye")


def initialize_connection(identifier: str, config_location: str or None) -> Connection:  # pragma: no cover
    """
    Loads the connection for the specified identifier
    If the connection was not found in the local configuration, the program exits.

    :param identifier: The identifier for the Connection
    :param config_location: The config location to use. Will default to the default value if None is supplied
    :return: The Connection object
    """

    config_handler = GlobalConfigHandler() if config_location is None else GlobalConfigHandler(config_location)

    try:
        config_handler.validate_config_directory()
    except InvalidConfigException as e:
        print("Loading configuration failed:")
        print(str(e))
        sys.exit(1)

    connections = config_handler.load_connections()
    services = config_handler.load_services()

    try:
        connection_type = list(filter(lambda x: x.define_identifier() == identifier, connections))[0]
        return connection_type(services, config_handler)
    except IndexError:
        print("Connection Type " + identifier + " is not implemented or imported using the config file")
        sys.exit(1)
    except InvalidConfigException as e:
        print("Connection Configuration failed:")
        print(str(e))
        sys.exit(1)


def parse_args() -> argparse.Namespace:  # pragma: no cover
    """
    Parses the Command Line Arguments using argparse

    :return: The parsed arguments
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("connection", help="The Type of Connection to use")

    parser.add_argument("-v", "--verbose", action="store_true", help="Activates verbose output")
    parser.add_argument("-d", "--debug", action="store_true", help="Activates debug-level logging output")
    parser.add_argument("-c", "--config", nargs="?", help="Overrides the configuration directory location")

    return parser.parse_args()



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


class StandardConfigWriter(object):
    """
    A config handler class that writes the standard configuration for connections
    and services into the config.
    """

    def __init__(self, config_handler: GlobalConfigHandler = GlobalConfigHandler()):
        """
        Initializes the Standard Config Writer's configuration file locations
        :param config_handler: The GlobalConfigHandler to use. Default to the default one.
        """
        self.connection_config = config_handler.global_connection_config_location
        self.service_config = config_handler.services_config_location

    def write_standard_connection_config(self):
        """
        Writes the standard connection configuration file

        :return: None
        """

        with open(self.connection_config, 'w') as config:
            for connection in \
                    ["from kudubot.connections.cli.CliConnection import CliConnection",
                     "from kudubot.connections.whatsapp.WhatsappConnection import WhatsappConnection",
                     "from kudubot.connections.telegram.TelegramConnection import TelegramConnection"]:
                config.write(connection + "\n")

    def write_standard_service_config(self):
        """
        Writes the standard service configuration file

        :return: None
        """

        with open(self.service_config, 'w') as config:
            for service in \
                    ["from kudubot.services.simple_responder.SimpleResponderService import SimpleResponderService",
                     "from kudubot.services.reminder.ReminderService import ReminderService",
                     "from kudubot.services.anime_reminder.AnimeReminderService import AnimeReminderService",
                     "from kudubot.services.jokes.JokesService import JokesService",
                     "from kudubot.services.external.hello_rust.HelloRustService import HelloRustService"]:
                config.write(service + "\n")



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import re
import time
import logging
import datetime
from threading import Thread
from typing import List, Dict
from kudubot.entities.Message import Message
from kudubot.services.HelperService import HelperService
from kudubot.services.reminder.database import initialize_database, store_reminder, get_unsent_reminders,\
    mark_reminder_sent, convert_datetime_to_string


class ReminderService(HelperService):
    """
    Class that implements a Service for the Kudubot framework that allows
    users to store reminder message that are then sent at a later time
    """

    def init(self):
        """
        Initializes the database table and starts a background thread that perpetually
        searches for expired reminders
        """
        self.logger.info("Initializing Reminder Database Table")
        initialize_database(self.connection.db)

        self.logger.info("Starting Reminder background thread")
        background = Thread(target=self.background_loop)
        background.daemon = True
        background.start()

    @staticmethod
    def define_requirements() -> List[str]:
        """
        Defines the dependencies for the Service

        :return: A list of dependencies
        """
        return []

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier for this service

        :return: The Service's identifier
        """
        return "reminder"

    def define_command_name(self, language: str):
        """
        Defines the command name for this service
        :param language: the language in which to get the command name
        :return: The command name in the specified language
        """
        return {"en": "/remind", "de": "/erinner"}[language]

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        :return: A dictionary used to translate any user-facing messages
        """
        return {
            "@remind_command": {"en": "/remind", "de": "/erinner"},
            "@list_argument": {"en": "list", "de": "auflisten"},
            "@second_singular": {"en": "second", "de": "sekunde"},
            "@second_plural": {"en": "seconds", "de": "sekunden"},
            "@minute_singular": {"en": "minute", "de": "minute"},
            "@minute_plural": {"en": "minutes", "de": "minuten"},
            "@hour_singular": {"en": "hour", "de": "stunde"},
            "@hour_plural": {"en": "hours", "de": "stunden"},
            "@day_singular": {"en": "day", "de": "tag"},
            "@day_plural": {"en": "days", "de": "tage"},
            "@week_singular": {"en": "week", "de": "woche"},
            "@week_plural": {"en": "weeks", "de": "wochen"},
            "@year_singular": {"en": "year", "de": "jahr"},
            "@year_plural": {"en": "years", "de": "jahre"},
            "@stored_reply_title": {"en": "Message Stored", "de": "Nachricht gespeichert"},
            "@stored_confirmation_message": {"en": "The reminder message has successfully been stored.",
                                             "de": "Die Erinnerungsnachricht wurde erfolgreich gespeichert"},
            "@list_stored_message_start": {"en": "The following reminders are pending",
                                           "de": "Die folgenden Erinnerungen stehen noch aus"},
            "@list_response_title": {"en": "List of Reminders", "de": "Liste der Erinnerungen"},
            "@invalid_title": {"en": "Invalid Reminder", "de": "Ungültige Erinnerung"},
            "@invalid_message": {"en": "The reminder is invalid. Please consult /remind syntax for more information.",
                                 "de": "Die Erinnerung ist fehlerhaft. Sehe dir /erinner syntax an, und versuch es "
                                       "dann noch einmal."}
        }

    def handle_message(self, message: Message):
        """
        Handles a message received by the Service

        :param message: The message to handle
        :return: None
        """
        super().handle_message(message)
        if not self.is_applicable_to_without_help_or_syntax(message):
            return

        target = message.get_direct_response_contact()
        command = self.parse_message(message.message_body.lower().strip(), self.determine_language(message))

        if command["mode"] == "store":
            store_reminder(self.connection.db,
                           command["data"]["message"],
                           command["data"]["due_time"],
                           target.database_id)
            self.reply_translated("@stored_reply_title", "@stored_confirmation_message", message)

        elif command["mode"] == "list":
            reminders = list(filter(lambda x: x["receiver"].database_id == target.database_id,
                                    get_unsent_reminders(self.connection.db)))
            text = "@list_stored_message_start:\n\n"
            for reminder in reminders:
                text += convert_datetime_to_string(reminder["due_time"]) + ":" + reminder["message"] + "\n"
            self.reply_translated("@list_response_title", text, message)

        elif command["mode"] == "invalid":
            self.reply_translated("@invalid_title", "@invalid_message", message)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if the Service is applicable to a message

        :param message: The message to check
        :return: True if the Service is applicable, otherwise False
        """
        if super().is_applicable_to(message):
            return True

        language = self.determine_language(message)
        body = message.message_body.lower().strip()

        regex = "^@remind_command ([0-9]+ (" \
                "@second_singular|@second_plural|@minute_singular|@minute_plural|@hour_singular|@hour_plural|" \
                "@day_singular|@day_plural|@week_singular|@week_plural|@year_singular|@year_plural) )+" \
                "\"[^\"]+\"$"
        lang_regex = re.compile(self.translate(regex, language))

        if re.match(lang_regex, body) or body == self.translate("@remind_command @list_argument", language):
            self.logger.info("Message is applicable")
            return True

        else:
            self.logger.info("Message is not applicable")
            return False

    def background_loop(self):
        """
        Perpetually checks for expiring reminders.

        :return: None
        """
        db = self.connection.get_database_connection_copy()
        while True:
            unsent = get_unsent_reminders(db)

            for reminder in unsent:
                if reminder["due_time"].timestamp() < datetime.datetime.utcnow().timestamp():
                    self.connection.send_message(Message("Reminder", reminder["message"],
                                                         reminder["receiver"], self.connection.user_contact))
                    mark_reminder_sent(db, reminder["id"])

            time.sleep(1)

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the syntax with which the user can interact with this service

        :param language: The language to use
        :return: The syntax description in the specified language
        """
        return {
            "en": "/remind X second(s) \"Message\"\n"
                  "/remind X minute(s) \"Message\"\n"
                  "/remind X hour(s) \"Message\"\n"
                  "/remind X day(s) \"Message\"\n"
                  "/remind X week(s) \"Message\"\n"
                  "/remind X year(s) \"Message\"\n"
                  "/remind tomorrow \"Message\"\n"
                  "/remind next week \"Message\"\n"
                  "/remind next year \"Message\"\n\n"
                  "Combinations:\n"
                  "/remind X hours Y minutes Z seconds \"Message\"\n",
            "de": "/erinner X sekunde(n) \"Message\"\n"
                  "/erinner X minute(n) \"Message\"\n"
                  "/erinner X stunde(n) \"Message\"\n"
                  "/erinner X tag(e) \"Message\"\n"
                  "/erinner X woche(n) \"Message\"\n"
                  "/erinner X jahr(e) \"Message\"\n"
                  "/erinner morgen \"Message\"\n"
                  "/erinner nächste woche \"Message\"\n"
                  "/erinner nächstes jahr \"Message\"\n"
                  "Kombinationen:\n"
                  "/erinner X stunden Y minuten Z sekunden \"Message\"\n"
        }[language]

    def determine_language(self, message: Message) -> str:
        """
        Determines the language used in a message

        :param message: The message to analyse
        :return: The language that was found
        """

        if message.message_body.startswith("/erinner"):
            return "de"
        else:
            return "en"

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message for this service in various languages

        :param language: The language to be used
        :return: The help description in the specified language
        """
        return {
            "en": "The reminder service allows you to store a reminder on the server to be "
                  "sent back to you at a specified time. See /remind syntax for possible ways "
                  "to use the reminder service.",
            "de": "Der Erinnerungsdienst erlaubt es einem Nutzer eine Erinnerung auf dem Server "
                  "zu speicher, welcher dann zu einem spezifizierten Zeitpunkt zurückgesendet wird. "
                  "Sehe dir /erinner syntax an, um dir ein Bild davon zu machen wie man den Erinnerungsdienst "
                  "verwendet."
        }[language]

    def parse_message(self, text: str, language: str) -> Dict[str, str or Dict[str, str or datetime]]:
        """
        Parses the message and determines the mode of operation

        :param text: The text to parse
        :param language: The language in which to parse the text
        :return: A dictionary with at least the key 'status' with three different possible states:
                 - no-match: The message does not match the command syntax
                 - help:     A query for the help message. Will result in the help message being sent to the sender
                 - store:    Command to store a new reminder
        """
        self.logger.debug("Parsing message")

        if self.translate("@remind_command @list_argument", language) == text:
            return {"mode": "list"}
        else:

            time_string = text.split(" \"")[0].split(self.translate("@remind_command ", language))[1]
            reminder_message = text.split("\"")[1]

            usertime = self.parse_time_string(time_string.strip(), language)

            if usertime is None:
                self.logger.debug("Invalid reminder message")
                return {"mode": "invalid"}
            else:
                self.logger.debug("Will store reminder")
                return {"mode": "store", "data": {"message": reminder_message, "due_time": usertime}}

    def parse_time_string(self, time_string: str, language: str) -> datetime or None:
        """
        Parses a time string like '1 week' or '2 weeks 1 day' etc. and returns
        a datetime object with the specified time difference to the current time.

        :param time_string: The time string to parse
        :param language: In which language the string should be parsed
        :return: The parsed datetime object or None in case the parsing failed
        """

        # turn all units into english singular units
        for unit in ["second", "minute", "hour", "day", "week", "year"]:
            for mode in ["plural", "singular"]:  # The order is very important here!

                key = "@" + unit + "_" + mode
                text = self.translate(key, language)
                time_string = time_string.replace(text, unit)

        now = datetime.datetime.utcnow()
        parsed = time_string.split(" ")

        try:
            for i in range(0, len(parsed), 2):

                value = int(parsed[i])
                key = parsed[i + 1]

                if key == "year":
                    now = now.replace(year=now.year + value)
                elif key == "month":
                    now = now.replace(month=now.month + value)
                elif key == "day":
                    now = now.replace(day=now.day + value)
                elif key == "hour":
                    now = now.replace(hour=now.hour + value)
                elif key == "minute":
                    now = now.replace(minute=now.minute + value)
                elif key == "second":
                    now = now.replace(second=now.second + value)
                else:
                    self.logger.debug("Invalid time keyword used")
                    return None

            return now

        except ValueError:  # Datetime exception
            return None



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

import os
import shutil
import logging
import importlib
from typing import List
from kudubot.services.Service import Service
from kudubot.exceptions import InvalidConfigException


class GlobalConfigHandler(object):
    """
    Class that handles the global kudubot configuration files located in $HOME/.kudubot
    """

    logger = logging.getLogger("kudubot.config.GlobalConfigHandler")
    """
    The Logger for this class
    """

    def __init__(self, config_location: str = os.path.join(os.path.expanduser("~"), ".kudubot")):
        """
        Initializes the ConfigHandler. Determines the config locations using the config_location
        parameter which defaults to a .kudubot directory in the user's home directory.
        The configuration may still be invalid once the object is initialized, call
        validate_config_directory() to make sure that the configuration is correct.

        :param config_location: The location of the config directory
        """
        self.config_location = config_location
        self.global_connection_config_location = os.path.join(self.config_location, "connections.conf")
        self.services_config_location = os.path.join(self.config_location, "services.conf")
        self.data_location = os.path.join(self.config_location, "data")
        self.specific_connection_config_location = os.path.join(self.config_location, "connection_config")
        self.external_services_directory = os.path.join(self.config_location, "external")
        self.external_services_executables_directory = os.path.join(self.external_services_directory, "bin")

    def validate_config_directory(self) -> bool:
        """
        Validates the configuration directory. As soon as a discrepancy is detected, the reason is logged
        and False is returned. If the configuration is valid however, True is returned
        :return: True if the config is valid, False otherwise
        """

        try:
            if not os.path.isdir(self.config_location):
                raise InvalidConfigException("Configuration directory " + self.config_location + " does not exist")
            elif not os.path.isfile(self.global_connection_config_location):
                raise InvalidConfigException("Connection config file does not exist")
            elif not os.path.isfile(self.services_config_location):
                raise InvalidConfigException("Services config file does not exist")
            elif not os.path.isdir(self.data_location):
                raise InvalidConfigException("Data Location directory does not exist")
            elif not os.path.isdir(self.specific_connection_config_location):
                raise InvalidConfigException("Connection Configuration directory does not exist")
            elif not os.path.isdir(self.external_services_directory):
                raise InvalidConfigException("External Service directory does not exist")
            elif not os.path.isdir(self.external_services_executables_directory):
                raise InvalidConfigException("External Service executable directory does not exist")
            else:
                self.logger.info("Configuration successfully checked")
                return True

        except InvalidConfigException as e:
            self.logger.warning("Configuration invalid: " + e.args[0])
            return False

    def generate_configuration(self, delete_old=False):
        """
        Generates a new, empty config location.

        :param delete_old: If set, all old config files that may exist are overwritten
        :return: None
        """

        if delete_old and os.path.isdir(self.config_location):
            self.logger.info("Deleting old configuration files")
            shutil.rmtree(self.config_location)

        if not os.path.isdir(self.config_location):
            self.logger.info("Creating directory " + self.config_location)
            os.makedirs(self.config_location)

        if not os.path.isfile(self.global_connection_config_location):
            self.logger.info("Creating file " + self.global_connection_config_location)
            open(self.global_connection_config_location, "w").close()

        if not os.path.isfile(self.services_config_location):
            self.logger.info("Creating file " + self.services_config_location)
            open(self.services_config_location, "w").close()

        if not os.path.isdir(self.data_location):
            self.logger.info("Creating directory " + self.data_location)
            os.makedirs(self.data_location)

        if not os.path.isdir(self.specific_connection_config_location):
            self.logger.info("Creating directory " + self.specific_connection_config_location)
            os.makedirs(self.specific_connection_config_location)

        if not os.path.isdir(self.external_services_directory):
            self.logger.info("Creating directory " + self.external_services_directory)
            os.makedirs(self.external_services_directory)

        if not os.path.isdir(self.external_services_executables_directory):
            self.logger.info("Creating directory " + self.external_services_executables_directory)
            os.makedirs(self.external_services_executables_directory)

    def load_connections(self) -> List[type]:
        """
        Loads all connections from the connections configuration file

        :return: A list of successfully imported Connection subclasses
        """
        from kudubot.connections.Connection import Connection

        self.logger.info("Loading connections")
        connections = self.__load_import_config__(self.global_connection_config_location, Connection)

        if len(connections) == 0:
            self.logger.warning("No connections loaded")

        return self.__remove_duplicate_services_or_connections__(connections)

    # noinspection PyUnresolvedReferences
    def load_services(self) -> List[type]:
        """
        Loads all Services from the services configuration file

        :return: A list of successfully imported Service subclasses
        """
        self.logger.info("Loading Services")
        services = self.__load_import_config__(self.services_config_location, Service)

        # Check if dependencies for each service are satisfied
        # This loop structure is admittedly a bit weird, but there's a valid reason!
        # Since we are removing Services with unsatisfied dependencies, we need to re-check all
        # previous Services again if a Service is removed.
        # This is done by directly modifying the index variable i
        i = 0
        while i < len(services):

            service = services[i]

            for dependency in service.define_requirements():

                dependency_satisfied = False
                for other_service in services:
                    if other_service.define_identifier() == dependency:
                        dependency_satisfied = True
                        break

                if not dependency_satisfied:
                    self.logger.warning(
                        "Dependency '" + dependency + "' for service '" +
                        service.define_identifier() + "' is not satisfied")
                    services.remove(service)
                    i = -1
                    break
            i += 1

        if len(services) == 0:
            self.logger.warning("No services loaded")

        return self.__remove_duplicate_services_or_connections__(services)

    # noinspection PyMethodMayBeStatic
    def __handle_import_statement__(self, statement: str) -> type:
        """
        Handles an import statement string

        :param statement: The import string to parse and execute
        :return: The retrieved class or module
        """

        if statement.startswith("import"):
            return importlib.import_module(statement.split("import ", 1)[1])
        else:
            statement = statement.split("from ", 1)[1]
            statement = statement.split(" import ")
            module = importlib.import_module(statement[0])
            return getattr(module, statement[1])

    # noinspection PyUnresolvedReferences,PyMethodMayBeStatic
    def __remove_duplicate_services_or_connections__(self, target: List[type]) -> List[type]:
        """
        Removes any duplicate Connections or Services from a list

        :param target: The list to remove the duplicates from
        :return: The list with duplicates removed
        """

        results = []

        for element in target:

            hitcount = 0

            for other in target:
                if other.define_identifier() == element.define_identifier():
                    hitcount += 1

            if hitcount == 1:
                results.append(element)
            else:

                exists = False

                for result in results:
                    if result.define_identifier() == element.define_identifier():
                        exists = True
                        break

                if not exists:
                    results.append(element)

        return results

    # noinspection PyMethodMayBeStatic
    def __load_import_config__(self, file_location: str, class_type: type) -> List[type]:
        """
        Reads an import config file (A file containing only python import statements)
        and returns a list of the successful imports.

        Imports that fail are simply ignored, as well as those that do not return
        subclasses of the class_type parameter.

        :param file_location: The import config file's location
        :param class_type: The class type all imports must subclass
        :return: The list of successful imports
        """

        modules = []

        with open(file_location, 'r') as config:
            content = config.read().split("\n")

        for line in content:

            self.logger.debug("Trying to import '" + line + "'")

            if line.strip().startswith("#") or line == "":
                self.logger.debug("Skipping line: " + line + "")
                continue
            else:
                try:
                    _module = self.__handle_import_statement__(line)

                    if issubclass(_module, class_type):
                        modules.append(_module)
                        self.logger.info("Import " + line + " successful")
                    else:
                        self.logger.warning("Import " + line + " is not of type " + str(class_type))

                except ImportError:  # Ignore invalid imports
                    self.logger.warning("Import " + line + " has failed")

        return modules

    def delete_service_executables(self):
        """
        Deletes all executable service files

        :return: None
        """
        shutil.rmtree(self.external_services_executables_directory)
        os.makedirs(self.external_services_executables_directory)



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""


import re
import time
import logging
from threading import Thread
from typing import List, Dict
from kudubot.entities.Message import Message
from kudubot.connections.Connection import Connection
from kudubot.services.HelperService import HelperService
from kudubot.services.anime_reminder.scraper import scrape_reddit_discussion_threads
from kudubot.services.anime_reminder.database import initialize_database, store_subscription, delete_subscription, \
    get_subscriptions, thread_exists, store_thread


class AnimeReminderService(HelperService):
    """
    The Kudubot Service that provides the anime reminder functionality
    """

    def init(self):
        """
        In addition to the normal initialization of a Service, this service initializes
        its database and starts the background thread
        """
        initialize_database(self.connection.db)

        self.logger.info("Starting Anime Reminder Background Thread")
        background = Thread(target=self.background_loop)
        background.daemon = True
        background.start()

    @staticmethod
    def define_requirements() -> List[str]:
        """
        :return: The dependencies for this service
        """
        return []

    @staticmethod
    def define_identifier() -> str:
        """
        :return: The service's identifier
        """
        return "anime_reminder"

    def handle_message(self, message: Message):
        """
        Handles a message from the user

        :param message: The message to handle
        :return: None
        """
        super().handle_message(message)
        if not self.is_applicable_to_without_help_or_syntax(message):
            return

        self.logger.debug("Handling message: " + message.message_body)

        mode = message.message_body.split(" ")[1].lower()
        user = message.get_direct_response_contact()
        user_id = user.database_id

        params = message.message_body.split("\"")
        show = "" if len(params) < 2 else params[1]

        if mode in [self.translate("@subscribe_command", x) for x in self.supported_languages()]:

            result = "@successful_add_message" if store_subscription(user_id, show, self.connection.db) \
                else "@exists_message"
            self.reply_translated("@reply_title", "@subscribe_message_start " + show + " " + result + ".", message)

        elif mode in [self.translate("@unsubscribe_command", x) for x in self.supported_languages()]:

            result = "@successful_remove_message" if delete_subscription(user_id, show, self.connection.db) \
                else "@does_not_exist_message"
            self.reply_translated("@reply_title", "@unsubscribe_message_start " + show + " " + result + ".", message)

        elif mode in [self.translate("@list_command", x) for x in self.supported_languages()]:

            message_text = "@subscription_list_message:\n\n"
            for subscription in get_subscriptions(self.connection.db, user_id):
                message_text += subscription["show_name"] + "\n"
            self.reply_translated("@reply_title", message_text, message)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if a message is applicable to this Service

        :param message: The message to check
        :return: None
        """

        if super().is_applicable_to(message):
            return True

        applicable = False
        for language in self.supported_languages():

            regex = "^@command_name (@list_command|@sub_unsub_command \"[^\"]+\")$"
            regex = re.compile(self.translate(regex, language))
            applicable = bool(re.search(regex, message.message_body)) or applicable

        self.logger.debug("Message " + message.message_body + " is" + ("" if applicable else "not") +
                          " applicable to " + self.define_identifier())

        return applicable

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the syntax description for the Service
        :param language: The language in which to display the syntax description
        :return: The syntax description in the specified language
        """
        skeleton = "@command_name @list_command\n"
        skeleton += "@command_name @subscribe_command \"@show_name_parameter\"\n"
        skeleton += "@command_name @unsubscribe_command \"@show_name_parameter\""
        return self.translate(skeleton, language)

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        Defines the Language strings for the service for the implemented languages
        :return: The dictionary used to translate strings
        """
        return {
            "@list_command": {"en": "list",
                              "de": "auflisten"},
            "@subscribe_command": {"en": "subscribe",
                                   "de": "abonnieren"},
            "@unsubscribe_command": {"en": "unsubscribe",
                                     "de": "abbestellen"},
            "@sub_unsub_command": {"en": "(un)?subscribe",
                                   "de": "(abonnieren|abbestellen)"},
            "@show_name_parameter": {"en": "Show Name",
                                     "de": "Serienname"},
            "@command_name": {"en": "/anime-remind",
                              "de": "/anime-erinner"},
            "@exists_message": {"en": "already exists",
                                "de": "existiert bereits"},
            '@does_not_exist_message': {"en": "does not exist",
                                        "de": "existiert nicht"},
            '@successful_add_message': {"en": "successful",
                                        "de": "wurde erfolgreich abonniert"},
            '@successful_remove_message': {"en": "successfully removed",
                                           "de": "wurde erfolgreich abbestellt"},
            "@subscribe_message_start": {"en": "Subscription for",
                                         "de": "Abonnement für"},
            "@unsubscribe_message_start": {"en": "Subscription for",
                                           "de": "Abonnement für"},
            "@reply_title": {"en": "Anime Reminder",
                             "de": "Anime Erinnerung"},
            "@remind_title": {"en": "Anime Reminder",
                              "de": "Anime Erinnerung"},
            "@subscription_list_message": {"en": "You are subscribed to the following shows",
                                           "de": "Du hast folgende Serien abonniert"},
            "@episode": {"en": "episode",
                         "de": "Episode"},
            "@has_been_released": {"en": "has been released! Discuss it on reddit",
                                   "de": "wurde veröffentlicht! Diskuttier sie auf reddit"},
            "@help_message": {
                "en": "The Anime Reminder Service periodically checks for new anime discussion threads on reddit's "
                      "r/anime board. A user can subscribe to specific shows and will then be notified whenever a "
                      "new thread appears, which usually coincides with the point in time that the episode is "
                      "available on streaming services like Crunchyroll, Amazon or Funimation.\n"
                      "Users can also unsubscribe from shows and list all their subscriptions. Subscription names "
                      "are case-insensitive.",
                "de": "Der Anime Erinnerungs-Service schaut in periodischen Abständen nach, ob neue Anime Diskussionen "
                      "auf reddit.com/r/anime entstanden sind, welche neue Anime Episode diskuttieren. Ein Nutzer kann "
                      "einzelne Serien abonnieren und wieder abbestellen. Für abonnierte Serien erhält der Nutzer "
                      "eine Nachricht sobal eine neue Episode verfügbar ist."
                      "Die Namen für die Abonnements achten nicht auf Groß- und Kleinschreibung. Man kann auch alle"
                      "Abonnements die derzeit aktiv sind auflisten lassen."
            }
        }

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message for the service

        :param language: The target language to translate the message to
        :return: The help message in the specified language
        """
        return self.translate("@help_message", language)

    def determine_language(self, message: Message) -> str:
        """
        Determines the language used based on the incoming message

        :param message: The message to analyze
        :return: The language key to use
        """
        if message.message_body.startswith(self.define_command_name("de")):
            return "de"
        else:
            return "en"

    def define_command_name(self, language: str) -> str:
        """
        Defines the command name for the anime-remind functionality

        :param language: The language for which the command is valid
        :return: The command name in the specified language
        """
        return self.translate("@command_name", language)

    def background_loop(self):
        """
        Starts a new thread which will continuously check for new anime discussion threads on reddit.com/r/anime
        and notify users if they are subscribed to those shows

        :return: None
        """
        db = self.connection.get_database_connection_copy()
        while True:
            new_threads = scrape_reddit_discussion_threads()

            self.logger.info("Checking for due subscriptions")

            for thread in new_threads:

                if not thread_exists(thread, db):
                    store_thread(thread, db)

                    for subscription in get_subscriptions(db):
                        if subscription["show_name"].lower() == thread["show_name"].lower():

                            receiver = self.connection.address_book.get_contact_for_id(subscription["user_id"], db)

                            message_text = thread["show_name"] + " @episode " + str(thread["episode"])
                            message_text += " @has_been_released: " + thread["url"]

                            language = self.get_language_preference(receiver.database_id, db=db)
                            message_text = self.translate(message_text, language)
                            title = self.translate("@remind_title", language)

                            self.connection.send_message(Message(title, message_text, receiver,
                                                                 self.connection.user_contact))

            time.sleep(60)



"""
LICENSE:
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

    kudubot is a chat bot framework. It allows developers to write
    services for arbitrary chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from typing import List

from kudubot.entities.Message import Message
from kudubot.services.Service import Service


class DummyService(Service):
    """
    A class that implements a Service for use in unit tests
    """

    @staticmethod
    def define_requirements() -> List[str]:
        return []

    @staticmethod
    def define_identifier() -> str:
        return "dummyservice"

    def handle_message(self, message: Message):
        pass

    def is_applicable_to(self, message: Message) -> bool:
        pass


class DummyServiceWithValidDependency(DummyService):
    """
    A Service that has itself (or DummyService) as its only dependency
    """

    @staticmethod
    def define_requirements() -> List[str]:
        return ["dummyservice"]


class DummyServiceWithInvalidDependency(DummyService):
    """
    A service that has an invalid dependency
    """

    @staticmethod
    def define_requirements() -> List[str]:
        return ["otherservice"]
